#ifndef CAFFE_BASE_CONV_HASH_LAYER_HPP_
#define CAFFE_BASE_CONV_HASH_LAYER_HPP_

#include <vector>

#include "caffe/blob.hpp"
#include "caffe/layer.hpp"
#include "caffe/proto/caffe.pb.h"
#include "caffe/util/HashData.h"
namespace caffe {

/**
 * @brief Abstract base class that factors out the BLAS code common to
 *        ConvolutionLayer and DeconvolutionLayer.
 */
template <typename Dtype>
class BaseConvHashLayer : public Layer<Dtype> {
 public:
  explicit BaseConvHashLayer(const LayerParameter& param)
	  : Layer<Dtype>(param) { conv_on_self_ = true; is_1x1_ = false; }
  virtual ~BaseConvHashLayer();
  virtual void LayerSetUp(const vector<Blob<Dtype>*>& bottom,
      const vector<Blob<Dtype>*>& top);
  virtual void Reshape(const vector<Blob<Dtype>*>& bottom,
      const vector<Blob<Dtype>*>& top);

  virtual inline int MinBottomBlobs() const { return 1; }
  virtual inline int MinTopBlobs() const { return 1; }
  virtual inline bool EqualNumBottomTopBlobs() const { return false; }

 protected:
	 void forward_cpu_gemm(const Dtype* input,
		 const Dtype* weights, Dtype* output);
  void forward_cpu_gemm_self(const Dtype* input,
		 const Dtype* weights, Dtype* output);		//for conv stride = 1, where bottom struct = top struct
  void forward_cpu_gemm_ds(const Dtype* input,
	  const Dtype* weights, Dtype* output);			//convolution stride>1
  void forward_cpu_bias(Dtype* output, const Dtype* bias);
  void backward_cpu_gemm(const Dtype* output,
	  const Dtype* weights, Dtype* input);
  void backward_cpu_gemm_self(const Dtype* output,
	  const Dtype* weights, Dtype* input);
  void backward_cpu_gemm_ds(const Dtype* output,
	  const Dtype* weights, Dtype* input);
  void weight_cpu_gemm(const Dtype* input,
	  const Dtype* output, Dtype* weights);
  void weight_cpu_gemm_self(const Dtype* input,
	  const Dtype* output, Dtype* weights);
  void weight_cpu_gemm_ds(const Dtype* input,
	  const Dtype* output, Dtype* weights);
  void backward_cpu_bias(Dtype* bias, const Dtype* input);

#ifndef CPU_ONLY
  void forward_gpu_gemm(const Dtype* input,
	  const Dtype* weights, Dtype* output);
  void forward_gpu_gemm_self(const Dtype* input,
	  const Dtype* weights, Dtype* output);
  void forward_gpu_gemm_ds(const Dtype* input,
	  const Dtype* weights, Dtype* output);
  void forward_gpu_bias(Dtype* output, const Dtype* bias);
  void backward_gpu_gemm(const Dtype* output,
	  const Dtype* weights, Dtype* input);
  void backward_gpu_gemm_self(const Dtype* output,
	  const Dtype* weights, Dtype* input);
  void backward_gpu_gemm_ds(const Dtype* output,
	  const Dtype* weights, Dtype* input);
  void weight_gpu_gemm(const Dtype* input,
	  const Dtype* output, Dtype* weights);
  void weight_gpu_gemm_self(const Dtype* input,
	  const Dtype* output, Dtype* weights);
  void weight_gpu_gemm_ds(const Dtype* input,
	  const Dtype* output, Dtype* weights);
  void backward_gpu_bias(Dtype* bias, const Dtype* input);
#endif

  // reverse_dimensions should return true iff we are implementing deconv, so
  // that conv helpers know which dimensions are which.
  virtual bool reverse_dimensions() = 0;
  // Compute height_out_ and width_out_ from other parameters.
  virtual void compute_output_shape() = 0;

  void reshape_topHashData(const vector<Blob<Dtype>*>& bottom,
	  const vector<Blob<Dtype>*>& top);
  void reshape_colBuf(const vector<Blob<Dtype>*>& bottom, const vector<Blob<Dtype>*>& top);
  /// @brief The spatial dimensions of a filter kernel.
  Blob<int> kernel_shape_;
  /// @brief The spatial dimensions of the stride.
  Blob<int> stride_;
  /// @brief The spatial dimensions of the padding.
  Blob<int> pad_;
  /// @brief The spatial dimensions of the dilation.
  Blob<int> dilation_;
  /// @brief The spatial dimensions of the convolution input.
  Blob<int> conv_input_shape_;
  /// @brief The spatial dimensions of the col_buffer.
  vector<int> col_buffer_shape_;
  /// @brief The spatial dimensions of the output.
  vector<int> output_shape_;

  int num_spatial_axes_;	//3 in 3D case
  int channels_;
  int num_output_;
  bool bias_term_;
  
  int conv_out_channels_;
  int conv_in_channels_;
  int kernel_dim_;
  int conv_out_spatial_dim_;
  int out_spatial_dim_;

  bool conv_on_self_;
  bool is_1x1_;

  int m_bottom_struct_idx;
  int m_top_struct_idx;

  //for conv
  int m_conv_bt_struct_idx;
  int m_conv_top_struct_idx;
public://for debug
	int writeDenseKernel_2_HF5(const char *filename);
	int writeBias_2_HF5(const char *filename);
	int writeDenseKernelDif_2_HF5(const char *filename);
	int writeBiasDif_2_HF5(const char *filename);

#ifdef USE_CUDNN	//to accelerate weight_gpu_gemm()
	// Interestingly, we find the GEMM is nearly ten times slower than Cudnn
	// when doing weight back-propagation. It's possible because the ill matrix
	// size (suach as A(1000, 120000) * B(120000, 500) = W(1000, 500)).
	// cudnn
	cudnnConvolutionBwdFilterAlgo_t bwd_filter_algo_;
	cudnnHandle_t                   handle_;
	cudnnTensorDescriptor_t         bottom_desc_;
	cudnnTensorDescriptor_t         top_desc_;
	cudnnFilterDescriptor_t         filter_desc_;
	cudnnConvolutionDescriptor_t    conv_desc_;
	size_t  filter_workspace_size_;
	Blob<Dtype> filter_workspace_;
#endif // USE_CUDNN

	//for acceleration
#if RECORD_NEIGHBOR
	Blob<int> nei_idx_;	//to accelerate bp
#endif
};


}  // namespace caffe

#endif  // CAFFE_BASE_CONVOLUTION_LAYER_HPP_
