#include <algorithm>
#include <cfloat>
#include <vector>

#include "caffe/common.hpp"
#include "caffe/layer.hpp"
#include "caffe/syncedmem.hpp"
#include "caffe/util/math_functions.hpp"
#include "caffe/layers/hash_2_dense_layer.hpp"

namespace caffe {

using std::min;
using std::max;

template <typename Dtype>
void Hash2DenseLayer<Dtype>::LayerSetUp(const vector<Blob<Dtype>*>& bottom,
	const vector<Blob<Dtype>*>& top) 
{
	//CHECK size
	if (bottom.size() != HASH_DATA_SIZE)	//data + input struct
	{
		printf("Fatal error: bottom size should be %d\n", HASH_DATA_SIZE);
		exit(0);
	}
	if (top.size() != 1)
	{
		printf("Fatal error: top size should be 1\n");
		exit(0);
	}


	num_spatial_axes_ = 3;	//for 3D case
	channels_ = bottom[HASH_DATA_BLOB]->shape(CHANNEL_AXIS);
	CHECK_GT(channels_, 0);

	const Hash2DenseParameter &h2d_param = this->layer_param_.hash2dense_param();
	m_struct_idx = h2d_param.bottom_struct();
}


template <typename Dtype>
void Hash2DenseLayer<Dtype>::Reshape(const vector<Blob<Dtype>*>& bottom,
	const vector<Blob<Dtype>*>& top)
{
	CBatchStructure self_struct;
	CGlobalSys<Dtype>::setBatchStructure(self_struct, m_struct_idx);	//top is the same as bottom

	int dense_res = self_struct.m_res_ptr[0];
	std::vector<int> top_shape;
	top_shape.push_back(self_struct.m_batch_size);	//N
	top_shape.push_back(channels_);	//C
	for (int i = 0; i<num_spatial_axes_; i++)	//spatial D H W
	{
		top_shape.push_back(dense_res);
	}
	top[0]->Reshape(top_shape);
}

template <typename Dtype>
void Hash2DenseLayer<Dtype>::Forward_cpu(const vector<Blob<Dtype>*>& bottom,
	const vector<Blob<Dtype>*>& top) 
{
	CBatchStructure self_struct;
	CGlobalSys<Dtype>::setBatchStructure(self_struct, m_struct_idx);	//top is the same as bottom

	//init 2 zero
	caffe_set(top[0]->count(),(Dtype)0., top[0]->mutable_cpu_data());

	batch_hash_2_dense(bottom[HASH_DATA_BLOB]->cpu_data(), &self_struct, channels_, top[HASH_DATA_BLOB]->mutable_cpu_data());
}


template <typename Dtype>
void Hash2DenseLayer<Dtype>::Backward_cpu(const vector<Blob<Dtype>*>& top,
	const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom) 
{
	CBatchStructure self_struct;
	CGlobalSys<Dtype>::setBatchStructure(self_struct, m_struct_idx);	//top is the same as bottom
											
	batch_dense_2_hash(bottom[HASH_DATA_BLOB]->mutable_cpu_diff(), &self_struct, channels_, top[HASH_DATA_BLOB]->cpu_diff());
}

#ifdef CPU_ONLY
STUB_GPU(Hash2DenseLayer);
#endif


INSTANTIATE_CLASS(Hash2DenseLayer);
REGISTER_LAYER_CLASS(Hash2Dense);
}  // namespace caffe
