#include <algorithm>
#include <cfloat>
#include <vector>


#include "caffe/layer.hpp"
#include "caffe/util/math_functions.hpp"
#include "caffe/layers/hash_2_dense_layer.hpp"
#include "caffe\util\HashData.h"

namespace caffe {
	template<class Dtype>
	__global__ void hash_2_dense_kernel(const Dtype* valid_data,
		const CBatchStructure self_struct,
		Dtype *dense_data, 
		int total_m, int channels)
	{
		CUDA_KERNEL_LOOP(cur_thread, channels * total_m)
		{
			const int total_defNum = self_struct.m_defNumSum_ptr[self_struct.m_batch_size];

			int C = cur_thread / total_m;
			int M_idx = cur_thread - C * total_m;

			const int valid_idx = self_struct.m_hashIdxTable_ptr[M_idx];	//valid idx in the hash within the batch
			const VolumeIndexType volIdx = self_struct.m_volIdx_ptr[M_idx];	//the idx of the hash in the batch
			const int defNumSum = (int)self_struct.m_defNumSum_ptr[volIdx];	//defNumSum before this hash
			const int dense_res = (int)self_struct.m_res_ptr[volIdx];	//dense resolution
			
			if (valid_idx < 0)
			{
				return;	//invalid hash slot, empty
			}

		
			const int global_valid_idx = defNumSum + valid_idx;
			const int data_idx = C * total_defNum + global_valid_idx;

			//get corresponding dense position
			const int3 voxel = xyz_from_pack_g(self_struct.m_posTag_ptr[M_idx]);
			const int res2 = dense_res * dense_res;
			const int res3 = res2 * dense_res;
			dense_data[(volIdx * channels + C) * res3 + voxel.z * res2 + voxel.y * dense_res + voxel.x] = valid_data[data_idx];
		} // end for caffe_kernel_loop
	}

	template <typename Dtype>
	void Hash2DenseLayer<Dtype>::Forward_gpu(const vector<Blob<Dtype>*>& bottom,
		const vector<Blob<Dtype>*>& top)
	{
		//struct GPU
		CBatchStructure bottom_struct_GPU;
		CGlobalSys<Dtype>::setBatchStructure_GPU(bottom_struct_GPU, m_struct_idx);

		CBatchStructure bottom_struct;
		CGlobalSys<Dtype>::setBatchStructure(bottom_struct, m_struct_idx);
	
		//init 2 zero
		caffe_gpu_set(top[0]->count(), (Dtype)0., top[0]->mutable_gpu_data());
	
		int total_m = (int)bottom_struct.m_mSum_ptr[bottom_struct.m_batch_size];

		hash_2_dense_kernel<Dtype> << <CAFFE_GET_BLOCKS(total_m*channels_), CAFFE_CUDA_NUM_THREADS >> > (bottom[HASH_DATA_BLOB]->gpu_data(),
			bottom_struct_GPU,
			top[HASH_DATA_BLOB]->mutable_gpu_data(),
			total_m, channels_);
	}


	template<class Dtype>
	__global__ void dense_2_hash_kernel(Dtype* valid_data,
		const CBatchStructure self_struct,
		const Dtype *dense_data,
		int total_m, int channels)
	{
		CUDA_KERNEL_LOOP(cur_thread, channels * total_m)
		{
			const int total_defNum = self_struct.m_defNumSum_ptr[self_struct.m_batch_size];

			int C = cur_thread / total_m;
			int M_idx = cur_thread - C * total_m;

			const int valid_idx = self_struct.m_hashIdxTable_ptr[M_idx];	//valid idx in the hash within the batch
			const VolumeIndexType volIdx = self_struct.m_volIdx_ptr[M_idx];	//the idx of the hash in the batch
			const int defNumSum = (int)self_struct.m_defNumSum_ptr[volIdx];	//defNumSum before this hash
			const int dense_res = (int)self_struct.m_res_ptr[volIdx];	//dense resolution

			if (valid_idx < 0)
			{
				return;	//invalid hash slot, empty
			}


			const int global_valid_idx = defNumSum + valid_idx;
			const int data_idx = C * total_defNum + global_valid_idx;

			//get corresponding dense position
			const int3 voxel = xyz_from_pack_g(self_struct.m_posTag_ptr[M_idx]);
			const int res2 = dense_res * dense_res;
			const int res3 = res2 * dense_res;
			valid_data[data_idx] = dense_data[(volIdx * channels + C) * res3 + voxel.z * res2 + voxel.y * dense_res + voxel.x];
		} // end for caffe_kernel_loop
	}


	template <typename Dtype>
	void Hash2DenseLayer<Dtype>::Backward_gpu(const vector<Blob<Dtype>*>& top,
		const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom)
	{
		//struct GPU
		CBatchStructure bottom_struct_GPU;
		CGlobalSys<Dtype>::setBatchStructure_GPU(bottom_struct_GPU, m_struct_idx);

		CBatchStructure bottom_struct;
		CGlobalSys<Dtype>::setBatchStructure(bottom_struct, m_struct_idx);


		int total_m = (int)bottom_struct.m_mSum_ptr[bottom_struct.m_batch_size];

		dense_2_hash_kernel<Dtype> << <CAFFE_GET_BLOCKS(total_m*channels_), CAFFE_CUDA_NUM_THREADS >> > (bottom[HASH_DATA_BLOB]->mutable_gpu_diff(),
			bottom_struct_GPU,
			top[HASH_DATA_BLOB]->gpu_diff(),
			total_m, channels_);
	}

	INSTANTIATE_LAYER_GPU_FUNCS(Hash2DenseLayer);
}  // namespace caffe
