#include <algorithm>
#include <cfloat>
#include <vector>

#include "caffe/common.hpp"
#include "caffe/layer.hpp"
#include "caffe/syncedmem.hpp"
#include "caffe/util/math_functions.hpp"
#include "caffe/layers/pool_hash_layer.hpp"

namespace caffe {

using std::min;
using std::max;

template <typename Dtype>
void PoolHashLayer<Dtype>::LayerSetUp(const vector<Blob<Dtype>*>& bottom,
	const vector<Blob<Dtype>*>& top) 
{
	//CHECK size
	if (bottom.size() != HASH_DATA_SIZE)
	{
		printf("Fatal error: bottom size should be %d\n", HASH_DATA_SIZE);
		exit(0);
	}
	if (top.size() != HASH_DATA_SIZE && top.size() != HASH_DATA_SIZE + 1)	//the "1" is for output mask
	{
		printf("Fatal error: top size should be %d or  %d\n", HASH_DATA_SIZE, HASH_DATA_SIZE+1);
		exit(0);
	}


	num_spatial_axes_ = 3;	//for 3D case

	channels_ = bottom[HASH_DATA_BLOB]->shape(CHANNEL_AXIS);

	const PoolHashParameter &pool_hash_param = this->layer_param_.pool_hash_param();
	m_bottom_struct_idx = pool_hash_param.bottom_struct();
	m_top_struct_idx = pool_hash_param.top_struct();

	// Configure the kernel size, padding, stride, and inputs.
	const vector<int> spatial_dim_blob_shape(1, num_spatial_axes_);
	
	// Setup filter kernel dimensions(kernel_shape_).
	kernel_shape_.Reshape(spatial_dim_blob_shape);
	int* kernel_shape_data = kernel_shape_.mutable_cpu_data();
	{
		const int num_kernel_dims = pool_hash_param.kernel_size_size();
		CHECK(num_kernel_dims == 1 || num_kernel_dims == num_spatial_axes_)
			<< "kernel_size must be specified once, or once per spatial dimension "
			<< "(kernel_size specified " << num_kernel_dims << " times; "
			<< num_spatial_axes_ << " spatial dims).";
		for (int i = 0; i < num_spatial_axes_; ++i) {
			kernel_shape_data[i] =
				pool_hash_param.kernel_size((num_kernel_dims == 1) ? 0 : i);
		}
	}
	for (int i = 0; i < num_spatial_axes_; ++i) {
		CHECK_GT(kernel_shape_data[i], 0) << "Filter dimensions must be nonzero.";
	}

	// Setup filter pad dimensions(pad_shape_).
	pad_shape_.Reshape(spatial_dim_blob_shape);
	for (int i = 0; i < num_spatial_axes_; i++)
		pad_shape_.mutable_cpu_data()[i] = (pool_hash_param.pad_size() == 0 ? 0 :
			pool_hash_param.pad((pool_hash_param.pad_size() == 1) ? 0 : i));
	

	// Setup filter stride dimensions(stride_shape_).
	stride_shape_.Reshape(spatial_dim_blob_shape);
	for (int i = 0; i < num_spatial_axes_; i++)
		stride_shape_.mutable_cpu_data()[i] = (pool_hash_param.stride_size() == 0 ? 1 :
			pool_hash_param.stride((pool_hash_param.stride_size() == 1) ? 0 : i));

	//Currently we only support kernel_size==stride, because in pooling_backward, we avoid atomicAdd to fairly compare speed with OCNN
	if (stride_shape_.cpu_data()[0] != kernel_shape_.cpu_data()[0])
	{
		LOG(INFO) << "Currently we only support kernel_size==stride, because in pooling_backward, we avoid atomicAdd to fairly compare speed with OCNN";
		exit(0);
	}
}



template <typename Dtype>
void PoolHashLayer<Dtype>::reshape_topHashData(const vector<Blob<Dtype>*>& bottom,
	const vector<Blob<Dtype>*>& top)
{
	const Blob<int>* top_defNumSum_blob = CGlobalSys<Dtype>::get_defNumSum_blob(m_top_struct_idx);
	const int total_top_defNum = top_defNumSum_blob->cpu_data()[top_defNumSum_blob->count() - 1];
	std::vector<int> data_shape;
#if ADD_NUM_2_SHAPE
	data_shape.push_back(1);
#endif
	data_shape.push_back(channels_);
	data_shape.push_back(total_top_defNum);
	top[HASH_DATA_BLOB]->Reshape(data_shape);


	//also reshape mask
	if (top.size() == HASH_DATA_SIZE + 1)
		top[HASH_DATA_SIZE]->ReshapeLike(*top[HASH_DATA_BLOB]);
	if (this->layer_param_.pooling_param().pool() == PoolingParameter_PoolMethod_MAX && top.size() == HASH_DATA_SIZE)
	{
		max_idx_.Reshape(data_shape);
	}	
}

template <typename Dtype>
void PoolHashLayer<Dtype>::Reshape(const vector<Blob<Dtype>*>& bottom,
	const vector<Blob<Dtype>*>& top)
{
	reshape_topHashData(bottom, top);
}

template <typename Dtype>
void PoolHashLayer<Dtype>::forward_cpu_max(const Dtype *bottom_valid_data, 
	const CBatchStructure *bottom_struct,
	Dtype *top_valid_data, const CBatchStructure *top_struct,
	int *mask, Dtype *top_mask)
{
	const int batch_size = top_struct->m_batch_size;
	const int top_total_defNum = top_struct->m_defNumSum_ptr[batch_size];
	const int bottom_total_defNum = bottom_struct->m_defNumSum_ptr[batch_size];
	const int channels = channels_;
	//init mask and top data
	if (mask) 
	{
		caffe_set(channels*top_total_defNum, -1, mask);
	}
	else {
		caffe_set(channels*top_total_defNum, Dtype(-1), top_mask);
	}
	caffe_set(channels*top_total_defNum, Dtype(-FLT_MAX), top_valid_data);

	
	const int stride = stride_shape_.cpu_data()[0];
	const int pad = pad_shape_.cpu_data()[0];
	const int kernel = kernel_shape_.cpu_data()[0];

	
	//starting bottom struct
	const unsigned char *bt_offset = bottom_struct->m_offset_ptr;
	const PACKED_POSITION *bt_posTag = bottom_struct->m_posTag_ptr;
	const int *bt_hash_table = bottom_struct->m_hashIdxTable_ptr;
	//starint top struct
	const PACKED_POSITION *top_posTag = top_struct->m_posTag_ptr;
	const int *tp_hash_table = top_struct->m_hashIdxTable_ptr;

	for (int b = 0; b < batch_size; b++)
	{
		//bottom struct
		const int bt_dense_res = (int)bottom_struct->m_res_ptr[b];
		const int bt_m_bar = (int)bottom_struct->m_mBar_ptr[b];
		const int bt_r_bar = (int)bottom_struct->m_rBar_ptr[b];
		const int bt_m = bt_m_bar * bt_m_bar * bt_m_bar;
		const int bt_r = bt_r_bar * bt_r_bar * bt_r_bar;
		const int bt_r2 = bt_r_bar * bt_r_bar;
		const int bt_m2 = bt_m_bar * bt_m_bar;
		const int bt_defNum = (int)bottom_struct->m_defNum_ptr[b];
		const int bt_defNumSum = (int)bottom_struct->m_defNumSum_ptr[b];

		//top struct
		const int top_m_bar = (int)top_struct->m_mBar_ptr[b];
		const int top_m = top_m_bar * top_m_bar * top_m_bar;
		const int top_m2 = top_m_bar * top_m_bar;
		const int top_defNum = (int)top_struct->m_defNum_ptr[b];
		const int top_defNumSum = (int)top_struct->m_defNumSum_ptr[b];
		for (int v = 0; v < top_m; v++)
		{
			//if the hash voxel is undefined, skip
			if (!ishashVoxelDefined(&top_posTag[v]))
			{
				continue;
			}
			int tp_valid_idx = tp_hash_table[v];
			int global_tp_valid_idx = top_defNumSum + tp_valid_idx;
			///////////////////////////////////////////

			//get the real voxel position from the position tag
			int cx, cy, cz;
			xyz_from_pack(top_posTag[v], cx, cy, cz);	//get the voxel position mapped to this hash														

			int min_x = cx * stride - pad;
			int min_y = cy * stride - pad;
			int min_z = cz * stride - pad;

			int x_end = min(min_x + kernel, bt_dense_res);
			int y_end = min(min_y + kernel, bt_dense_res);
			int z_end = min(min_z + kernel, bt_dense_res);

			min_x = max(min_x, 0);
			min_y = max(min_y, 0);
			min_z = max(min_z, 0);

			int bt_mx, bt_my, bt_mz;

			for (int nz = min_z; nz < z_end; ++nz)
			{
				for (int ny = min_y; ny < y_end; ++ny)
				{
					for (int nx = min_x; nx < x_end; ++nx)
					{
						//hash to get hash position
						Hash(nx, ny, nz, bt_mx, bt_my, bt_mz,
							bt_offset, bt_m_bar, bt_r_bar, bt_r2);
						const int bt_m_idx = NXYZ2I(bt_mx, bt_my, bt_mz, bt_m_bar, bt_m2);

						if (!ishashVoxelDefined(&bt_posTag[bt_m_idx]))	//the bottom hash voxel is undefined
						{
							continue;
						}

						int stored_x, stored_y, stored_z;
						xyz_from_pack(bt_posTag[bt_m_idx], stored_x, stored_y, stored_z);
						if (nx != stored_x || ny != stored_y || nz != stored_z)	//undefined dense voxel
						{
							continue;
						}

						int bt_valid_idx = bt_hash_table[bt_m_idx];
						if (bt_valid_idx < 0 || bt_valid_idx >= bt_defNum)
						{
							printf("Fatal error: postag conflicts with hash index table!\n");
							exit(0);
						}
						//we record the global index in bottom_blob
						int global_bt_valid_idx = bt_defNumSum + bt_valid_idx;

						const Dtype *bt_ptr = &bottom_valid_data[global_bt_valid_idx];
						Dtype *tp_ptr = &top_valid_data[global_tp_valid_idx];
						int *mask_ptr = NULL;
						Dtype *top_mask_ptr = NULL;
						if (mask)
							mask_ptr = &mask[global_tp_valid_idx];
						else
							top_mask_ptr = &top_mask[global_tp_valid_idx];
						for (int c = 0; c < channels; c++)
						{
							if (*tp_ptr < *bt_ptr)
							{
								*tp_ptr = *bt_ptr;
								if(mask_ptr) 
									*mask_ptr = global_bt_valid_idx;
								else 
									*top_mask_ptr = (Dtype)global_bt_valid_idx;
							}
							tp_ptr += top_total_defNum;
							bt_ptr += bottom_total_defNum;
							if(mask_ptr) 
								mask_ptr += top_total_defNum;
							else 
								top_mask_ptr += top_total_defNum;
						}
					}
				}
			}
#if USE_EMPTY_VALID_REGION
			Dtype *tp_ptr = &top_valid_data[global_tp_valid_idx];
			//new added, for expanded empty valid regions, set to zero
			if (*tp_ptr == -FLT_MAX)
			{
				for (int c = 0; c<channels; c++)
				{
					*tp_ptr = 0;
					tp_ptr += top_total_defNum;
				}
			}
#endif
		}

		bt_offset += bt_r * 3;
		bt_posTag += bt_m;
		bt_hash_table += bt_m;

		top_posTag += top_m;
		tp_hash_table += top_m;
	}
}


template <typename Dtype>
void PoolHashLayer<Dtype>::Forward_cpu_max(const vector<Blob<Dtype>*>& bottom,
	const vector<Blob<Dtype>*>& top)
{
	//get bottom struct
	CBatchStructure bottom_struct;
	CGlobalSys<Dtype>::setBatchStructure(bottom_struct, m_bottom_struct_idx);	//top is the same as bottom
	//get top struct
	CBatchStructure top_struct;
	CGlobalSys<Dtype>::setBatchStructure(top_struct, m_top_struct_idx);	//top is the same as bottom


	bool use_top_mask = (top.size() == (HASH_DATA_SIZE + 1));
	int* mask = NULL;  // suppress warnings about uninitalized variables
	Dtype* top_mask = NULL;
	if (use_top_mask)
	{
		top_mask = top[HASH_DATA_SIZE]->mutable_cpu_data();
	}
	else
	{
		mask = max_idx_.mutable_cpu_data();
	}

	forward_cpu_max(bottom[HASH_DATA_BLOB]->cpu_data(),
		&bottom_struct,
		top[HASH_DATA_BLOB]->mutable_cpu_data(), &top_struct,
		mask, top_mask);

	//for debug
	//FILE *fp = fopen("max_idx_CPU.bin","wb");
	//if (use_top_mask)
	//{
	//	fwrite(top_mask, sizeof(Dtype), top[HASH_DATA_SIZE]->count(), fp);
	//}
	//else
	//{
	//	fwrite(mask, sizeof(int), max_idx_.count(), fp);
	//}
	//fclose(fp);


	//for debug
	//if (top_struct.m_res_ptr[0] == 256)
	//{
	//	static int counter;
	//	char buf[128];
	//	sprintf(buf, "debug_pool_forward_%02d_CPU.txt",counter++);
	//	std::ofstream debug_ofs(buf);
	//	debug_ofs << "===============Top data==========================" << std::endl;
	//	const Dtype *top_data = top[HASH_DATA_BLOB]->cpu_data();
	//	for (int i = 0; i < top[HASH_DATA_BLOB]->shape(SPATIAL_AXIS); i++)
	//	{
	//		const Dtype *cur_data = top_data;
	//		for (int j = 0; j < top[HASH_DATA_BLOB]->shape(CHANNEL_AXIS); j++)
	//		{
	//			debug_ofs << (int)*cur_data << " ";
	//			cur_data += top[HASH_DATA_BLOB]->shape(SPATIAL_AXIS);
	//		}
	//		debug_ofs << std::endl;
	//		top_data++;
	//	}
	//}
}


template <typename Dtype>
void PoolHashLayer<Dtype>::Forward_cpu(const vector<Blob<Dtype>*>& bottom,
	const vector<Blob<Dtype>*>& top) 
{
	switch (this->layer_param_.pooling_param().pool()) {
	case PoolingParameter_PoolMethod_MAX:
		Forward_cpu_max(bottom, top);
		break;
	case PoolingParameter_PoolMethod_AVE:
		NOT_IMPLEMENTED;
		break;
	case PoolingParameter_PoolMethod_STOCHASTIC:
		NOT_IMPLEMENTED;
		break;
	default:
		LOG(FATAL) << "Unknown pooling method.";
	}
}


template <typename Dtype>
void PoolHashLayer<Dtype>::backward_cpu_max(Dtype *bottom_valid_dif, const CBatchStructure *bottom_struct,
		const Dtype *top_valid_dif, const CBatchStructure *top_struct,
		const int *mask, const Dtype *top_mask)
{
	const int top_total_def_num = top_struct->m_defNumSum_ptr[top_struct->m_batch_size];
	const int bottom_total_def_num = bottom_struct->m_defNumSum_ptr[bottom_struct->m_batch_size];
	//init dif to zero
	caffe_set(bottom_total_def_num*channels_, (Dtype)0, bottom_valid_dif);
	for (int v = 0; v < top_total_def_num; v++)
	{
		const Dtype *tp_dif_ptr = &top_valid_dif[v];
		const int *mask_ptr = NULL;
		const Dtype *top_mask_ptr = NULL;
		if (mask)
			mask_ptr = &mask[v];
		else
			top_mask_ptr = &top_mask[v];

		Dtype *bt_dif_start = bottom_valid_dif;
		
#if USE_EMPTY_VALID_REGION
		if (mask_ptr && *mask_ptr == -1 ||
			top_mask_ptr && *top_mask_ptr==(Dtype)-1)	//expanded regions, no parent
		{

		}
		else
		{
			for (int c = 0; c < channels_; c++)
			{
				const int bt_idx = (mask ? *mask_ptr : (int)*top_mask_ptr);

				bt_dif_start[bt_idx] += *tp_dif_ptr;

				tp_dif_ptr += top_total_def_num;
				bt_dif_start += bottom_total_def_num;
				if (mask)
					mask_ptr += top_total_def_num;
				else
					top_mask_ptr += top_total_def_num;
			}
		}
#else
		for (int c = 0; c < channels_; c++)
		{
			const int bt_idx = (mask ? *mask_ptr : (int)*top_mask_ptr);
			
			bt_dif_start[bt_idx] += *tp_dif_ptr;

			tp_dif_ptr += top_total_def_num;
			bt_dif_start += bottom_total_def_num;
			if (mask)
				mask_ptr += top_total_def_num;
			else
				top_mask_ptr += top_total_def_num;
		}
#endif
	}
}



template <typename Dtype>
void PoolHashLayer<Dtype>::Backward_cpu_max(const vector<Blob<Dtype>*>& top,
	const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom) 
{
	//get bottom struct
	CBatchStructure bottom_struct;
	CGlobalSys<Dtype>::setBatchStructure(bottom_struct, m_bottom_struct_idx);	//top is the same as bottom
																				//get top struct
	CBatchStructure top_struct;
	CGlobalSys<Dtype>::setBatchStructure(top_struct, m_top_struct_idx);	//top is the same as bottom


	bool use_top_mask = (top.size() == (HASH_DATA_SIZE + 1));
	const int* mask = NULL;  // suppress warnings about uninitalized variables
	const Dtype* top_mask = NULL;
	if (use_top_mask)
	{
		top_mask = top[HASH_DATA_SIZE]->cpu_data();
	}
	else
	{
		mask = max_idx_.cpu_data();
	}

	backward_cpu_max(bottom[HASH_DATA_BLOB]->mutable_cpu_diff(),
		&bottom_struct,
		top[HASH_DATA_BLOB]->cpu_diff(), &top_struct,
		mask, top_mask);
}

template <typename Dtype>
void PoolHashLayer<Dtype>::Backward_cpu(const vector<Blob<Dtype>*>& top,
	const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom) 
{
	if (!propagate_down[HASH_DATA_BLOB]) {
		return;
	}
	switch (this->layer_param_.pooling_param().pool()) {
	case PoolingParameter_PoolMethod_MAX:
		Backward_cpu_max(top, propagate_down,bottom);
		break;
	case PoolingParameter_PoolMethod_AVE:
		NOT_IMPLEMENTED;
		break;
	case PoolingParameter_PoolMethod_STOCHASTIC:
		NOT_IMPLEMENTED;
		break;
	default:
		LOG(FATAL) << "Unknown pooling method.";
	}
}

#ifdef CPU_ONLY
STUB_GPU(PoolHashLayer);
#endif

INSTANTIATE_CLASS(PoolHashLayer);
REGISTER_LAYER_CLASS(PoolHash);
}  // namespace caffe


void bp_max_dense(const float *top_dif, const int *top_mask, float *bottom_dif, int top_res, int bottom_res, int channels)
{
	int top_n = top_res * top_res * top_res;
	int bottom_n = bottom_res * bottom_res * bottom_res;
	memset(bottom_dif,0,sizeof(float)*channels*bottom_n);

	for (int c=0;c<channels;c++)
	{
		const float *cur_top_dif = top_dif + top_n*c;
		float *cur_bottom_dif = bottom_dif + bottom_n*c;
		const int *cur_mask = top_mask + top_n*c;
		for (int ti=0;ti<top_n;ti++)
		{
			int bi = cur_mask[ti];
			if (bi<0)
			{
				continue;
			}
			cur_bottom_dif[bi] += cur_top_dif[ti];
		}
	}
}