#include <cfloat>
#include <vector>

#include "caffe/layers/scale_hash_layer.hpp"
#include "caffe/util/math_functions.hpp"
#include "caffe/util/HashData.h"

namespace caffe {

template <typename Dtype>
__global__ void TruncationLowerBounded_Hash_kernel(const int n, Dtype* in_out, Dtype lower_bound) {
  CUDA_KERNEL_LOOP(index, n) {
    in_out[index] = max(in_out[index], lower_bound);
  }
}

template <typename Dtype>
__global__ void TruncationUpperBounded_Hash_kernel(const int n, Dtype* in_out, Dtype upper_bound) {
  CUDA_KERNEL_LOOP(index, n) {
    in_out[index] = min(in_out[index], upper_bound);
  }
}


template<class Dtype>
__global__ void ScaleBiasForward_kernel(
	const Dtype *bottom_data, Dtype *top_data,
	int channels, int spatial_dim,
	const Dtype* scale, const Dtype* bias)
{
	CUDA_KERNEL_LOOP(threadId, spatial_dim*channels)
	{
		const int c = threadId / spatial_dim;
		top_data[threadId] = bottom_data[threadId] * scale[c] + bias[c];
	}
}

template<class Dtype>
__global__ void ScaleForward_kernel(
	const Dtype *bottom_data, Dtype *top_data,
	int channels, int spatial_dim,
	const Dtype* scale)
{
	CUDA_KERNEL_LOOP(threadId, spatial_dim*channels)
	{
		const int c = threadId / spatial_dim;
		top_data[threadId] = bottom_data[threadId] * scale[c];
	}
}


template <typename Dtype>
void ScaleHashLayer<Dtype>::Forward_gpu(const vector<Blob<Dtype>*>& bottom, const vector<Blob<Dtype>*>& top) 
{
  Dtype* scale_data = this->blobs_[0]->mutable_gpu_data();
  if (this->layer_param_.scale_hash_param().has_min_value()) {
    // NOLINT_NEXT_LINE(whitespace/operators)
    TruncationLowerBounded_Hash_kernel<Dtype> << <CAFFE_GET_BLOCKS(channels_), CAFFE_CUDA_NUM_THREADS >> >(
		channels_, scale_data, this->layer_param_.scale_hash_param().min_value());
    CUDA_POST_KERNEL_CHECK;
  }
  if (this->layer_param_.scale_hash_param().has_max_value()) {
    // NOLINT_NEXT_LINE(whitespace/operators)
    TruncationUpperBounded_Hash_kernel<Dtype> << <CAFFE_GET_BLOCKS(channels_), CAFFE_CUDA_NUM_THREADS >> >(
      channels_, scale_data, this->layer_param_.scale_hash_param().max_value());
    CUDA_POST_KERNEL_CHECK;
  }

  int spatial_dim = bottom[HASH_DATA_BLOB]->shape(SPATIAL_AXIS);

  if (this->layer_param_.scale_hash_param().bias_term())
  {
	  ScaleBiasForward_kernel << <CAFFE_GET_BLOCKS(spatial_dim*channels_), CAFFE_CUDA_NUM_THREADS >> > (
		  bottom[HASH_DATA_BLOB]->gpu_data(), top[HASH_DATA_BLOB]->mutable_gpu_data(),
		  channels_,spatial_dim,
		  this->blobs_[0]->gpu_data(), this->blobs_[1]->gpu_data()
		  );
  }
  else
  {
	  ScaleForward_kernel << <CAFFE_GET_BLOCKS(spatial_dim*channels_), CAFFE_CUDA_NUM_THREADS >> > (
		  bottom[HASH_DATA_BLOB]->gpu_data(), top[HASH_DATA_BLOB]->mutable_gpu_data(),
		  channels_, spatial_dim,
		  this->blobs_[0]->gpu_data()
		  );
  }
}


template <typename Dtype>
void ScaleHashLayer<Dtype>::Backward_gpu(const vector<Blob<Dtype>*>& top,
    const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom) 
{
	const int total_def_num = top[HASH_DATA_BLOB]->shape(SPATIAL_AXIS);
	const Dtype *top_dif = top[HASH_DATA_BLOB]->gpu_diff();
	Dtype *bottom_dif = bottom[HASH_DATA_BLOB]->mutable_gpu_diff();
	const Dtype* bottom_data = bottom[HASH_DATA_BLOB]->gpu_data();
	const Blob<Dtype> &multiplier = CGlobalSys<Dtype>::get_multiplier();

	//bias bp 
	if (this->layer_param_.scale_hash_param().bias_term() &&
		this->param_propagate_down_.back())
	{
		caffe_gpu_gemv<Dtype>(CblasNoTrans, channels_, total_def_num, 1.,
			top_dif, multiplier.gpu_data(), 0, this->blobs_[1]->mutable_gpu_diff());
	}
	//weight bp
	if (this->param_propagate_down_[0])
	{
		// Hack: store big eltwise product in bottom[0] diff, except in the special
		// case where this layer itself does the eltwise product, in which case we
		// can store it directly in the scale diff, and we're done.
		// If we're computing in-place (and not doing eltwise computation), this
		// hack doesn't work and we store the product in temp_.
		Dtype* product = bottom_dif;
		caffe_gpu_mul(bottom[HASH_DATA_BLOB]->count(), top_dif, bottom_data, product);

		caffe_gpu_gemv(CblasNoTrans, channels_, total_def_num,
			Dtype(1), product, multiplier.gpu_data(), Dtype(0),
			this->blobs_[0]->mutable_gpu_diff());
	}

	if (propagate_down[0])
	{
		ScaleForward_kernel << <CAFFE_GET_BLOCKS(total_def_num*channels_), CAFFE_CUDA_NUM_THREADS >> > (
			top_dif,bottom_dif,channels_,total_def_num, this->blobs_[0]->gpu_data()
			);
	}
}


INSTANTIATE_LAYER_GPU_FUNCS(ScaleHashLayer);

}  // namespace caffe
