#include <algorithm>
#include <cfloat>
#include <vector>


#include "caffe/layer.hpp"
#include "caffe/util/math_functions.hpp"
#include "caffe/layers/unpool_hash_layer.hpp"
#include "caffe\util\HashData.h"

namespace caffe {

	template<class Dtype>
	__global__ void unpoolhash_forward_gpu_max_kernel(
		const Dtype *bottom_data, int bottom_total_def_num,
		Dtype *top_data, int top_total_def_num,
		const Dtype *mask, const int channels)
	{
		CUDA_KERNEL_LOOP(threadId, bottom_total_def_num*channels)
		{
			const int c = threadId / bottom_total_def_num;
			const int global_tp_valid_idx = mask[threadId];
#if USE_EMPTY_VALID_REGION
			if (global_tp_valid_idx != -1)
				top_data[global_tp_valid_idx + c*top_total_def_num] = bottom_data[threadId];
#else
			top_data[global_tp_valid_idx + c*top_total_def_num] = bottom_data[threadId];
#endif
		}
	}


	template<class Dtype>
	__global__ void unpoolhash_backward_gpu_max_kernel(Dtype *bottom_diff, int bottom_total_def_num,
		const Dtype *top_diff, int top_total_def_num,
		const Dtype *mask, const int channels)
	{
		CUDA_KERNEL_LOOP(threadId, bottom_total_def_num*channels)
		{
			const int c = threadId / bottom_total_def_num;
			const int global_tp_valid_idx = mask[threadId];
#if USE_EMPTY_VALID_REGION
			if (global_tp_valid_idx != -1)
				bottom_diff[threadId] = top_diff[global_tp_valid_idx + c*top_total_def_num];
#else
			bottom_diff[threadId] = top_diff[global_tp_valid_idx + c*top_total_def_num];
#endif
		}
	}

	template <typename Dtype>
	void UnPoolHashLayer<Dtype>::Forward_gpu(const vector<Blob<Dtype>*>& bottom,
		const vector<Blob<Dtype>*>& top)
	{
		switch (this->layer_param_.pooling_param().pool()) {
		case PoolingParameter_PoolMethod_MAX:
			Forward_gpu_max(bottom, top);
			break;
		case PoolingParameter_PoolMethod_AVE:
			NOT_IMPLEMENTED;
			break;
		case PoolingParameter_PoolMethod_STOCHASTIC:
			NOT_IMPLEMENTED;
			break;
		default:
			LOG(FATAL) << "Unknown pooling method.";
		}
		//LOG(INFO) << "UnpoolHash Forward GPU done";
	}

	template <typename Dtype>
	void UnPoolHashLayer<Dtype>::Forward_gpu_max(const vector<Blob<Dtype>*>& bottom,
		const vector<Blob<Dtype>*>& top)
	{
		CBatchStructure bottom_struct_CPU;
		CGlobalSys<Dtype>::setBatchStructure(bottom_struct_CPU, m_bottom_struct_idx);	//top is the same as bottom
		const int bottom_total_def_num = bottom_struct_CPU.m_defNumSum_ptr[bottom_struct_CPU.m_batch_size];
		//get top struct
		CBatchStructure top_struct_CPU;
		CGlobalSys<Dtype>::setBatchStructure(top_struct_CPU, m_top_struct_idx);	//top is the same as bottom
		const int top_total_def_num = top_struct_CPU.m_defNumSum_ptr[top_struct_CPU.m_batch_size];
		//struct CPU


		const Dtype* mask = bottom[HASH_DATA_SIZE]->gpu_data();

		//init top data to zero
		caffe_gpu_set(top[HASH_DATA_BLOB]->count(), Dtype(0.), top[HASH_DATA_BLOB]->mutable_gpu_data());

		unpoolhash_forward_gpu_max_kernel<Dtype> << <CAFFE_GET_BLOCKS(channels_*bottom_total_def_num), CAFFE_CUDA_NUM_THREADS >> > (
			bottom[HASH_DATA_BLOB]->gpu_data(), bottom_total_def_num,
			top[HASH_DATA_BLOB]->mutable_gpu_data(), top_total_def_num,
			mask, channels_);
	}

	template <typename Dtype>
	void UnPoolHashLayer<Dtype>::Backward_gpu(const vector<Blob<Dtype>*>& top,
		const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom)
	{
		if (!propagate_down[HASH_DATA_BLOB]) {
			return;
		}
		//LOG(INFO) << "UnpoolHash Backward GPU";
		switch (this->layer_param_.pooling_param().pool()) {
		case PoolingParameter_PoolMethod_MAX:
			Backward_gpu_max(top, propagate_down, bottom);
			break;
		case PoolingParameter_PoolMethod_AVE:
			NOT_IMPLEMENTED;
			break;
		case PoolingParameter_PoolMethod_STOCHASTIC:
			NOT_IMPLEMENTED;
			break;
		default:
			LOG(FATAL) << "Unknown pooling method.";
		}
		//LOG(INFO) << "UnpoolHash Backward GPU done";
	}

	template <typename Dtype>
	void UnPoolHashLayer<Dtype>::Backward_gpu_max(const vector<Blob<Dtype>*>& top,
		const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom)
	{
		CBatchStructure bottom_struct_CPU;
		CGlobalSys<Dtype>::setBatchStructure(bottom_struct_CPU, m_bottom_struct_idx);	//top is the same as bottom
		const int bottom_total_def_num = bottom_struct_CPU.m_defNumSum_ptr[bottom_struct_CPU.m_batch_size];
		//get top struct
		CBatchStructure top_struct_CPU;
		CGlobalSys<Dtype>::setBatchStructure(top_struct_CPU, m_top_struct_idx);	//top is the same as bottom
		const int top_total_def_num = top_struct_CPU.m_defNumSum_ptr[top_struct_CPU.m_batch_size];
		//struct CPU


		const Dtype* mask = bottom[HASH_DATA_SIZE]->gpu_data();

		//init bottom dif to zero
		/****************************************/
		//currently we can comment this, because in the kernel we set valud for every bottom dif
		//caffe_gpu_set(bottom[HASH_DATA_BLOB]->count(), Dtype(0.), bottom[HASH_DATA_BLOB]->mutable_gpu_diff());
		/****************************************/
		unpoolhash_backward_gpu_max_kernel<Dtype> << <CAFFE_GET_BLOCKS(channels_*bottom_total_def_num), CAFFE_CUDA_NUM_THREADS >> > (
			bottom[HASH_DATA_BLOB]->mutable_gpu_diff(), bottom_total_def_num,
			top[HASH_DATA_BLOB]->gpu_diff(), top_total_def_num,
			mask, channels_);
	}

	INSTANTIATE_LAYER_GPU_FUNCS(UnPoolHashLayer);
}  // namespace caffe
